import pprint

from pycoshark.mongomodels import Project, IssueSystem, Issue, Event, VCSSystem, Tag, Commit


def get_issues_with_their_events(issue_system_id, last_release_date, this_release_date, this_release_branches):

    pipeline = [
        {
            "$match": {"issue_system_id": issue_system_id}
        },
        {
            "$lookup": {
                'from': "event",
                'localField': "_id",
                'foreignField': "issue_id",
                'as': "events"
            },
        },
        {
            "$lookup": {
                'from': "issue_comment",
                'localField': "_id",
                'foreignField': "issue_id",
                'as': "comments"
            },
        },
        {
            "$lookup": {
                'from': "commit",
                'localField': "events.commit_id",
                'foreignField': "_id",
                'as': "ref_commits"
            },
        },

    ]

    if this_release_branches:
        pipeline.append({
            "$match": {
                'events.status': 'closed',
                'events.commit_id': {'$exists': True},
                'ref_commits.committer_date': {'$gte': last_release_date, '$lte': this_release_date},
                'ref_commits.branches': {'$in': this_release_branches}
            }
        })
    else:
        pipeline.append({
            "$match": {
                'events.status': 'closed',
                'events.commit_id': {'$exists': True},
                'ref_commits.committer_date': {'$gte': last_release_date, '$lte': this_release_date},
            }
        })

    pipeline.append({
        '$sort': {'created_at': -1},
    })

    return list(Issue.objects.aggregate(*pipeline))

analyzed_versions = [
   ('commons-beanutils','BEANUTILS_1_9_3', 'BEANUTILS_1_8_3'),
   ('commons-codec', 'commons-codec-1.11', '1.10'),
   ('commons-collections', 'collections-4.1', 'collections-4.0'),
   ('commons-io', '2.5', '2.4'),
   ('commons-lang', 'LANG_3_7', 'LANG_3_6'),
   ('commons-math', 'MATH_3_6', 'MATH_3_5'),
   ('druid', '1.1.0', '1.0.1'),
   ('fastjson', '1.2.41', '1.1.20'),
   ('google-gson', 'gson-parent-2.8.0', 'gson-parent-2.7'),
   ('guice', '4.1', '4.0'),
   ('HikariCP', 'HikariCP-2.7.0', 'HikariCP-2.6.0'),
   ('jackson-core', 'jackson-core-2.9.0', 'jackson-core-2.8.0'),
   ('jfreechart', 'v1.5.0', 'v1.0.19'),
   ('joda-time', 'v2.9', 'v2.8'),
   ('jsoup', 'jsoup-1.11.1', 'jsoup-1.10.1'),
   ('mybatis-3', 'mybatis-3.4.0', 'mybatis-3.3.0'),
   ('zxing', 'zxing-3.3.0', 'zxing-3.2.0')
]

for project_name, project_version, last_tag in analyzed_versions:
    project_id = Project.objects(name=project_name).get().id
    vcs_system_id = VCSSystem.objects(project_id=project_id).get().id
    issue_system_id = IssueSystem.objects(project_id=project_id).get().id

    this_release = Tag.objects(vcs_system_id=vcs_system_id, name=project_version).get()
    this_release_commit = Commit.objects(id=this_release.commit_id).get()

    last_release = Tag.objects(vcs_system_id=vcs_system_id, name=last_tag).get()
    last_release_date = Commit.objects(id=last_release.commit_id).get().committer_date

    print("Looking for project %s (%s)..." % (project_name, project_version))
    print("Looking for issues between %s and %s..." % (this_release_commit.committer_date, last_release_date))
    retrieved_issues = get_issues_with_their_events(
        issue_system_id,
        last_release_date,
        this_release_commit.committer_date,
        this_release_commit.branches
    )

    # Get issue_id, referenced_commits, commit_messages
    for issue in retrieved_issues:
        print("------------------------------------------------------------------")
        print("Issue Number: #%s" % issue['external_id'])
        print("Title: %s" % issue['title'])
        print("Description: %s" % issue['desc'])

        print()
        print()
        print("------------Issue Comments------------")
        for comment in issue['comments']:
            print(comment['comment'])

        print()
        print()
        print("------------Commit Information------------")
        for commit in issue['ref_commits']:
            print("Commit Date: %s" % commit['committer_date'])
            print("Commit Message: %s" % commit['message'])
            print("Commit Hash: %s" % commit['revision_hash'])

            print()
            print()

        print("------------------------------------------------------------------")
